#!/usr/bin/env python

"""
Handlers for a person for whom scheduling is performed.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from email.mime.text import MIMEText
from imiptools.config import MANAGER_PATH, MANAGER_URL
from imiptools.content import Handler, get_address, get_uri, to_part, uri_dict, uri_items
from imiptools.handlers.common import CommonFreebusy
from socket import gethostname
from vCalendar import to_node

def get_manager_url():
    url_base = MANAGER_URL or "http://%s/" % gethostname()
    return "%s/%s" % (url_base.rstrip("/"), MANAGER_PATH.lstrip("/"))

def get_object_url(uid):
    return "%s/%s" % (get_manager_url().rstrip("/"), uid)

class PersonHandler(Handler):

    "Handling mechanisms specific to people."

    def _record_and_deliver(self, objtype, from_organiser=True, queue=False, cancel=False):

        oa = self.require_organiser_and_attendees(from_organiser)
        if not oa:
            return False

        (organiser, organiser_attr), attendees = organiser_item, attendees = oa

        # Validate the organiser or attendee, ignoring spoofed requests.

        if not self.validate_identities(from_organiser and [organiser_item] or attendees.items()):
            return False

        # Handle notifications and invitations.

        if from_organiser:

            # Process each attendee separately.

            for attendee, attendee_attr in attendees.items():

                if not self.have_new_object(attendee, objtype):
                    continue

                # Store the object and queue any request.

                self.store.set_event(attendee, self.uid, to_node(
                    {objtype : [(self.details, {})]}
                    ))

                if queue:
                    self.store.queue_request(attendee, self.uid)
                elif cancel:
                    self.store.cancel_event(attendee, self.uid)

        # As organiser, update attendance.

        else:
            obj = self.get_object(organiser, objtype)

            if obj and self.have_new_object(organiser, objtype, obj):

                # Get attendee details in a usable form.

                attendee_map = uri_dict(self.get_value_map("ATTENDEE"))

                for attendee, attendee_attr in attendees.items():

                    # Update attendance in the loaded object.

                    attendee_map[attendee] = attendee_attr

                    # NOTE: Record other attendee free/busy information.

                # Set the new details and store the object.

                obj["ATTENDEE"] = attendee_map.items()

                self.store.set_event(organiser, self.uid, to_node(
                    {objtype : [(obj, {})]}
                    ))

        return True

    def _record_freebusy(self, from_organiser=True):

        "Record free/busy information for the received information."

        freebusy = []

        for value in self.get_values("FREEBUSY") or []:
            if not isinstance(value, list):
                value = [value]
            for v in value:
                try:
                    start, end = v.split("/", 1)
                    freebusy.append((start, end))
                except ValueError:
                    pass

        for sender, sender_attr in uri_items(self.get_items(from_organiser and "ORGANIZER" or "ATTENDEE")):
            for recipient in self.recipients:
                self.store.set_freebusy_for_other(get_uri(recipient), freebusy, sender)

    def wrap(self, method, text, from_organiser=True, link=True):

        "Wrap any valid message and pass it on to the recipient."

        texts = []
        texts.append(text)
        if link:
            texts.append("If your mail program cannot handle this "
                         "message, you may view the details here:\n\n%s" %
                         get_object_url(self.uid))

        return method, MIMEText("\n".join(texts))

class Event(PersonHandler):

    "An event handler."

    def add(self):

        # NOTE: Queue a suggested modification to any active event.

        # The message is now wrapped and passed on to the recipient.

        return "ADD", MIMEText("An addition to an event has been received.")

    def cancel(self):

        "Queue a cancellation of any active event."

        self._record_and_deliver("VEVENT", from_organiser=True, queue=False, cancel=True)
        return self.wrap("CANCEL", "A cancellation has been received.", from_organiser=True, link=True)

    def counter(self):

        # NOTE: Queue a suggested modification to any active event.

        # The message is now wrapped and passed on to the recipient.

        return "COUNTER", MIMEText("A counter proposal has been received.")

    def declinecounter(self):

        # NOTE: Queue a suggested modification to any active event.

        # The message is now wrapped and passed on to the recipient.

        return "DECLINECOUNTER", MIMEText("A declining counter proposal has been received.")

    def publish(self):

        "Register details of any relevant event."

        self._record_and_deliver("VEVENT", from_organiser=True, queue=False)
        return self.wrap("PUBLISH", "Details of an event have been received.", from_organiser=True, link=True)

    def refresh(self):

        "Update details of any active event."

        self._record_and_deliver("VEVENT", from_organiser=True, queue=False)
        return self.wrap("REFRESH", "An event update has been received.", from_organiser=True, link=True)

    def reply(self):

        "Record replies and notify the recipient."

        self._record_and_deliver("VEVENT", from_organiser=False, queue=False)
        return self.wrap("REPLY", "A reply has been received.", from_organiser=False, link=True)

    def request(self):

        "Hold requests and notify the recipient."

        self._record_and_deliver("VEVENT", from_organiser=True, queue=True)

        # The message is now wrapped and passed on to the recipient.

        return "REQUEST", MIMEText("A request has been queued and can be viewed here: %s" % get_object_url(self.uid))

class Freebusy(PersonHandler, CommonFreebusy):

    "A free/busy handler."

    def publish(self):

        "Register free/busy information."

        # NOTE: This could be configured to not produce a message.

        self._record_freebusy(from_organiser=True)
        return self.wrap("PUBLISH", "A free/busy update has been received.", from_organiser=True, link=False)

    def reply(self):

        "Record replies and notify the recipient."

        # NOTE: This could be configured to not produce a message.

        self._record_freebusy(from_organiser=False)
        return self.wrap("REPLY", "A reply has been received.", from_organiser=False, link=False)

    def request(self):

        """
        Respond to a request by preparing a reply containing free/busy
        information for each indicated attendee.
        """

        # NOTE: This should be subject to policy/preferences.

        return CommonFreebusy.request(self)

class Journal(PersonHandler):

    "A journal entry handler."

    def add(self):

        # NOTE: Queue a suggested modification to any active entry.

        # The message is now wrapped and passed on to the recipient.

        return "ADD", MIMEText("An addition to a journal entry has been received.")

    def cancel(self):

        # NOTE: Queue a suggested modification to any active entry.

        # The message is now wrapped and passed on to the recipient.

        return "CANCEL", MIMEText("A cancellation has been received.")

    def publish(self):

        # NOTE: Register details of any relevant entry.

        # The message is now wrapped and passed on to the recipient.

        self._record_and_deliver("VJOURNAL", from_organiser=True, queue=False)
        return self.wrap("PUBLISH", "Details of a journal entry have been received.", from_organiser=True, link=False)

class Todo(PersonHandler):

    "A to-do item handler."

    def add(self):

        # NOTE: Queue a suggested modification to any active item.

        # The message is now wrapped and passed on to the recipient.

        return "ADD", MIMEText("An addition to an item has been received.")

    def cancel(self):

        # NOTE: Queue a suggested modification to any active item.

        # The message is now wrapped and passed on to the recipient.

        return "CANCEL", MIMEText("A cancellation has been received.")

    def counter(self):

        # NOTE: Queue a suggested modification to any active item.

        # The message is now wrapped and passed on to the recipient.

        return "COUNTER", MIMEText("A counter proposal has been received.")

    def declinecounter(self):

        # NOTE: Queue a suggested modification to any active item.

        # The message is now wrapped and passed on to the recipient.

        return "DECLINECOUNTER", MIMEText("A declining counter proposal has been received.")

    def publish(self):

        "Register details of any relevant item."

        self._record_and_deliver("VTODO", from_organiser=True, queue=False)
        return self.wrap("PUBLISH", "Details of an item have been received.", from_organiser=True, link=True)

    def refresh(self):

        "Update details of any active item."

        self._record_and_deliver("VTODO", from_organiser=True, queue=False)
        return self.wrap("REFRESH", "An item update has been received.", from_organiser=True, link=True)

    def reply(self):

        "Record replies and notify the recipient."

        self._record_and_deliver("VTODO", from_organiser=False, queue=False)
        return self.wrap("REPLY", "A reply has been received.", from_organiser=False, link=True)

    def request(self):

        "Hold requests and notify the recipient."

        self._record_and_deliver("VTODO", from_organiser=True, queue=True)

        # The message is now wrapped and passed on to the recipient.

        return "REQUEST", MIMEText("A request has been queued.")

# Handler registry.

handlers = [
    ("VFREEBUSY",   Freebusy),
    ("VEVENT",      Event),
    ("VTODO",       Todo),
    ("VJOURNAL",    Journal),
    ]

# vim: tabstop=4 expandtab shiftwidth=4
